from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Tests for closure cache class.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2017, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import unittest

import logging
log = logging.getLogger(__name__)

from django.conf    import settings

from annalist.models.site           import Site
from annalist.models.collection     import Collection
from annalist.models.closurecache   import ClosureCache, Closure_Error

from .AnnalistTestCase import AnnalistTestCase
from .tests import (
    TestHost, TestHostUri, TestBasePath, TestBaseUri, TestBaseDir
    )

#   -----------------------------------------------------------------------------
#
#   ClosureCache tests
#
#   -----------------------------------------------------------------------------

class ClosureCacheTest(AnnalistTestCase):
    """
    Tests for transitive closure calculations in ClosureCache class.
    """

    def setUp(self):
        self.testsite     = Site(TestBaseUri, TestBaseDir)
        self.testcoll     = Collection(self.testsite, "testcoll")
        self.closurecache = ClosureCache(self.testcoll.get_id(), "test:rel")
        return

    def tearDown(self):
        return

    def test_empty_closure(self):
        self.assertEqual(self.closurecache.get_collection_id(), "testcoll")
        self.assertEqual(self.closurecache.get_relation_uri(),  "test:rel")
        self.assertEqual(self.closurecache.get_values(),         set())
        self.assertEqual(self.closurecache.fwd_closure("val1"),  set())
        self.assertEqual(self.closurecache.rev_closure("val2"),  set())
        return

    def test_singleton_closure(self):
        self.assertTrue(self.closurecache.add_rel("val1", "val2"))
        self.assertEqual(self.closurecache.get_values(),        {"val1", "val2"})
        self.assertEqual(self.closurecache.fwd_closure("val1"), {"val2"})
        self.assertEqual(self.closurecache.rev_closure("val2"), {"val1"})
        return

    def test_empty_closure_2(self):
        self.assertTrue(self.closurecache.add_rel("val1", "val2"))
        self.assertFalse(self.closurecache.add_rel("val1", "val2"))
        self.closurecache.add_rel("val1", "val2")
        self.closurecache.remove_val("val1")
        self.assertEqual(self.closurecache.get_values(),        set())
        self.assertEqual(self.closurecache.fwd_closure("val1"), set())
        self.assertEqual(self.closurecache.rev_closure("val2"), set())
        return

    def test_empty_closure_3(self):
        self.assertTrue(self.closurecache.add_rel("val1", "val2"))
        self.assertFalse(self.closurecache.add_rel("val1", "val2"))
        self.assertTrue(self.closurecache.remove_val("val2"))
        self.assertFalse(self.closurecache.remove_val("val2"))
        self.assertEqual(self.closurecache.get_values(),        set())
        self.assertEqual(self.closurecache.fwd_closure("val1"), set())
        self.assertEqual(self.closurecache.rev_closure("val2"), set())
        return

    def test_simple_transitive_closure(self):
        self.assertTrue(self.closurecache.add_rel("val1", "val2"))
        self.assertTrue(self.closurecache.add_rel("val2", "val3"))
        self.assertEqual(self.closurecache.get_values(),        {"val1", "val2", "val3"})
        self.assertEqual(self.closurecache.fwd_closure("val1"), {"val2", "val3"})
        self.assertEqual(self.closurecache.fwd_closure("val2"), {"val3"})
        self.assertEqual(self.closurecache.fwd_closure("val3"), set())
        self.assertEqual(self.closurecache.rev_closure("val1"), set())
        self.assertEqual(self.closurecache.rev_closure("val2"), {"val1"})
        self.assertEqual(self.closurecache.rev_closure("val3"), {"val1", "val2"})
        return

    def test_extended_transitive_closure(self):
        self.assertTrue(self.closurecache.add_rel("val1", "val2"))
        self.assertTrue(self.closurecache.add_rel("val2", "val3"))
        self.assertTrue(self.closurecache.add_rel("val3", "val4"))
        self.assertTrue(self.closurecache.add_rel("val3", "val5"))
        self.assertEqual(self.closurecache.get_values(),        {"val1", "val2", "val3", "val4", "val5"})
        self.assertEqual(self.closurecache.fwd_closure("val1"), {"val2", "val3", "val4", "val5"})
        self.assertEqual(self.closurecache.fwd_closure("val2"), {"val3", "val4", "val5"})
        self.assertEqual(self.closurecache.fwd_closure("val3"), {"val4", "val5"})
        self.assertEqual(self.closurecache.fwd_closure("val4"), set())
        self.assertEqual(self.closurecache.fwd_closure("val5"), set())
        self.assertEqual(self.closurecache.rev_closure("val1"), set())
        self.assertEqual(self.closurecache.rev_closure("val2"), {"val1"})
        self.assertEqual(self.closurecache.rev_closure("val3"), {"val1", "val2"})
        self.assertEqual(self.closurecache.rev_closure("val4"), {"val1", "val2", "val3"})
        self.assertEqual(self.closurecache.rev_closure("val5"), {"val1", "val2", "val3"})
        return

    def test_multipath_transitive_closure(self):
        self.assertTrue(self.closurecache.add_rel("val1", "val2"))
        self.assertTrue(self.closurecache.add_rel("val2", "val3"))
        self.assertTrue(self.closurecache.add_rel("val2", "val4"))
        self.assertTrue(self.closurecache.add_rel("val3", "val5"))
        self.assertTrue(self.closurecache.add_rel("val4", "val5"))
        self.assertEqual(self.closurecache.get_values(),        {"val1", "val2", "val3", "val4", "val5"})
        self.assertEqual(self.closurecache.fwd_closure("val1"), {"val2", "val3", "val4", "val5"})
        self.assertEqual(self.closurecache.fwd_closure("val2"), {"val3", "val4", "val5"})
        self.assertEqual(self.closurecache.fwd_closure("val3"), {"val5"})
        self.assertEqual(self.closurecache.fwd_closure("val4"), {"val5"})
        self.assertEqual(self.closurecache.fwd_closure("val5"), set())
        self.assertEqual(self.closurecache.rev_closure("val1"), set())
        self.assertEqual(self.closurecache.rev_closure("val2"), {"val1"})
        self.assertEqual(self.closurecache.rev_closure("val3"), {"val1", "val2"})
        self.assertEqual(self.closurecache.rev_closure("val4"), {"val1", "val2"})
        self.assertEqual(self.closurecache.rev_closure("val5"), {"val1", "val2", "val3", "val4"})
        return

    def test_multipath_closure_remove_path(self):
        self.assertTrue(self.closurecache.add_rel("val1", "val2"))
        self.assertTrue(self.closurecache.add_rel("val2", "val3"))
        self.assertTrue(self.closurecache.add_rel("val2", "val4"))
        self.assertTrue(self.closurecache.add_rel("val3", "val5"))
        self.assertTrue(self.closurecache.add_rel("val4", "val5"))
        self.assertTrue(self.closurecache.remove_val("val4"))
        self.assertFalse(self.closurecache.remove_val("val4"))
        self.assertEqual(self.closurecache.get_values(),        {"val1", "val2", "val3", "val5"})
        self.assertEqual(self.closurecache.fwd_closure("val1"), {"val2", "val3", "val5"})
        self.assertEqual(self.closurecache.fwd_closure("val2"), {"val3", "val5"})
        self.assertEqual(self.closurecache.fwd_closure("val3"), {"val5"})
        self.assertEqual(self.closurecache.fwd_closure("val4"), set())
        self.assertEqual(self.closurecache.fwd_closure("val5"), set())
        self.assertEqual(self.closurecache.rev_closure("val1"), set())
        self.assertEqual(self.closurecache.rev_closure("val2"), {"val1"})
        self.assertEqual(self.closurecache.rev_closure("val3"), {"val1", "val2"})
        self.assertEqual(self.closurecache.rev_closure("val4"), set())
        self.assertEqual(self.closurecache.rev_closure("val5"), {"val1", "val2", "val3"})
        return

    def test_multipath_closure_remove_nexus(self):
        self.assertTrue(self.closurecache.add_rel("val1", "val2"))
        self.assertTrue(self.closurecache.add_rel("val2", "val3"))
        self.assertTrue(self.closurecache.add_rel("val2", "val4"))
        self.assertTrue(self.closurecache.add_rel("val3", "val5"))
        self.assertTrue(self.closurecache.add_rel("val4", "val5"))
        self.assertTrue(self.closurecache.remove_val("val2"))
        self.assertFalse(self.closurecache.remove_val("val2"))
        self.assertEqual(self.closurecache.get_values(),        {"val3", "val4", "val5"})
        self.assertEqual(self.closurecache.fwd_closure("val1"), set())
        self.assertEqual(self.closurecache.fwd_closure("val2"), set())
        self.assertEqual(self.closurecache.fwd_closure("val3"), {"val5"})
        self.assertEqual(self.closurecache.fwd_closure("val4"), {"val5"})
        self.assertEqual(self.closurecache.fwd_closure("val5"), set())
        self.assertEqual(self.closurecache.rev_closure("val1"), set())
        self.assertEqual(self.closurecache.rev_closure("val2"), set())
        self.assertEqual(self.closurecache.rev_closure("val3"), set())
        self.assertEqual(self.closurecache.rev_closure("val4"), set())
        self.assertEqual(self.closurecache.rev_closure("val5"), {"val3", "val4"})
        return

    def test_reflexive_relation(self):
        self.assertTrue(self.closurecache.add_rel("val1", "val2"))
        with self.assertRaises(Closure_Error):
            self.closurecache.add_rel("val1", "val1")
        with self.assertRaises(Closure_Error):
            self.closurecache.add_rel("val2", "val2")
        self.assertEqual(self.closurecache.get_values(), {"val1", "val2"})
        return

    def test_recursive_relation(self):
        self.assertTrue(self.closurecache.add_rel("val1", "val2"))
        self.assertTrue(self.closurecache.add_rel("val2", "val3"))
        self.assertTrue(self.closurecache.add_rel("val3", "val4"))
        self.assertTrue(self.closurecache.add_rel("val3", "val5"))
        with self.assertRaises(Closure_Error):
            self.closurecache.add_rel("val5", "val1")
        with self.assertRaises(Closure_Error):
            self.closurecache.add_rel("val3", "val2")
        self.assertEqual(self.closurecache.get_values(), {"val1", "val2", "val3", "val4", "val5"})
        return

# End.



"""
Initialize Annalist server data.
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import sys
import logging
import importlib
import subprocess

log = logging.getLogger(__name__)

from utils.SuppressLoggingContext   import SuppressLogging
from utils.py3porting               import bytes_to_str 

from annalist.util                  import ensure_dir

from .                              import am_errors
from .am_settings                   import am_get_settings

def am_initialize(annroot, userhome, userconfig, options):
    """
    Initialize Annalist server data, database, etc.

    annroot     is the root directory for the annalist software installation.
    userhome    is the home directory for the host system user issuing the initialize command.
    userconfig  is the directory used for user-specific configuration files.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings_obj = am_get_settings(annroot, userhome, options)
    if not settings_obj:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) != 0:
        print("Unexpected arguments for initialize: (%s)"%(" ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    # Get config base directory from settings, and make sure it exists
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings_obj.modulename)

    # For unknown reason, the database path in DATABASES gets zapped, 
    # so code changed to use separately saved DATABASE_PATH.
    providersdir = os.path.join(sitesettings.CONFIG_BASE, "providers")
    databasedir  = os.path.dirname(sitesettings.DATABASE_PATH)
    ensure_dir(providersdir)
    ensure_dir(databasedir)
    # Initialze the database
    status = am_errors.AM_SUCCESS
    subprocess_command = (
        "django-admin migrate --pythonpath=%s --settings=%s"%
        (annroot, settings_obj.modulename)
        )
    log.debug("am_initialize subprocess: %s"%subprocess_command)
    # status = subprocess.call(
    #     subprocess_command.split(), 
    #     #stdout=sys.stdout, stderr=sys.stderr
    #     )
    # Allow stdout and stderr to be captured for testing
    p = subprocess.Popen(
        subprocess_command.split(), 
        stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
    out, err = p.communicate(None)
    status = p.returncode
    sys.stdout.write(bytes_to_str(out))
    sys.stderr.write(bytes_to_str(err))
    log.debug("am_initialize subprocess status: %s"%status)
    return status

def am_collectstatic(annroot, userhome, userconfig, options):
    """
    Collect Annalist statis data to (e.g.) `annalist_site/static`

    annroot     is the root directory for the annalist software installation.
    userhome    is the home directory for the host system user issuing the initialize command.
    userconfig  is the directory used for user-specific configuration files.
    options     contains options parsed from the command line.

    returns     0 if all is well, or a non-zero status code.
                This value is intended to be used as an exit status code
                for the calling program.
    """
    settings_obj = am_get_settings(annroot, userhome, options)
    if not settings_obj:
        print("Settings not found (%s)"%(options.configuration), file=sys.stderr)
        return am_errors.AM_NOSETTINGS
    if len(options.args) != 0:
        print("Unexpected arguments for collectstatic: (%s)"%(" ".join(options.args)), file=sys.stderr)
        return am_errors.AM_UNEXPECTEDARGS
    # Get config base directory from settings, and make sure it exists
    with SuppressLogging(logging.INFO):
        sitesettings = importlib.import_module(settings_obj.modulename)

    # Assemble static data
    status = am_errors.AM_SUCCESS
    subprocess_command = (
        "django-admin collectstatic --pythonpath=%s --settings=%s --clear --noinput"%
        (annroot, settings_obj.modulename)
        )
    log.debug("am_collectstatic subprocess: %s"%subprocess_command)
    # Allow stdout and stderr to be captured for testing
    p = subprocess.Popen(
        subprocess_command.split(), 
        stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
    out, err = p.communicate(None)
    status = p.returncode
    sys.stdout.write(bytes_to_str(out))
    sys.stderr.write(bytes_to_str(err))
    log.debug("am_collectstatic subprocess status: %s"%status)
    return status

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Annalist user record

A user is represented in a collectionm by:
- an ID (slug)
- a URI (currently a mailto: URI)
- a label (full name)
- a description
- a list of permissions applicable to the collection
- ...

The ID and URI must be matched by the authenticated issuer of an HTTP request for the
permissionms to be applied.  Other fields are cosmetic.
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import os.path
import shutil

import logging
log = logging.getLogger(__name__)

from django.conf import settings

from annalist                   import layout
from annalist.exceptions        import Annalist_Error
from annalist.identifiers       import ANNAL
from annalist                   import util
from annalist.models.entity     import Entity
from annalist.models.entitydata import EntityData

site_default_user_id  = "_site_default_user_perms"
site_default_user_uri = "annal:User/_default_user_perms"

default_user_id  = "_default_user_perms"
default_user_uri = "annal:User/_default_user_perms"

unknown_user_id  = "_unknown_user_perms"
unknown_user_uri = "annal:User/_unknown_user_perms"

class AnnalistUser(EntityData):

    _entitytype     = ANNAL.CURIE.User
    _entitytypeid   = layout.USER_TYPEID
    _entityroot     = layout.COLL_USER_PATH
    _entityview     = layout.COLL_USER_VIEW
    _entityfile     = layout.USER_META_FILE

    def __init__(self, parent, user_id):
        """
        Initialize a new AnnalistUser object, without metadata (yet).

        parent      is the parent entity from which the type is descended.
        user_id     the local identifier for the user
        altparent   is a site object to search for this new entity,
                    allowing site-wide AnnalistUser values to be found.
        """
        super(AnnalistUser, self).__init__(parent, user_id)
        # log.debug("AnnalistUser %s: dir %s"%(user_id, self._entitydir))
        # log.debug("AnnalistUser %s: url %s, viewurl %s"%(user_id, self._entityurl, self._entityviewurl))
        return

    def _migrate_values(self, userpermissions):
        """
        User permission data format migration method.
        """
        migration_map = (
            [ (ANNAL.CURIE.user_permissions, ANNAL.CURIE.user_permission)
            ])
        userpermissions = self._migrate_values_map_field_names(migration_map, userpermissions)
        return userpermissions

# End.



"""
Annalist record list

A record type is represented by:
- an ID (slug)
- a URI
- a name/label
- a description
- ...
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import os
import os.path
import shutil

from django.conf import settings

from annalist                   import layout
from annalist.exceptions        import Annalist_Error
from annalist.identifiers       import ANNAL
from annalist                   import util
from annalist.models.entity     import Entity
from annalist.models.entitydata import EntityData
from annalist.util              import extract_entity_id, make_type_entity_id

class RecordList(EntityData):

    _entitytype     = ANNAL.CURIE.List
    _entitytypeid   = layout.LIST_TYPEID
    _entityroot     = layout.COLL_LIST_PATH
    _entityview     = layout.COLL_LIST_VIEW
    _entityfile     = layout.LIST_META_FILE

    def __init__(self, parent, list_id):
        """
        Initialize a new RecordList object, without metadta (yet).

        parent      is the parent collection in which the list is defined.
        list_id     the local identifier for the record list
        altparent   is a site object to search for this new entity,
                    allowing site-wide RecordType values to be found.
        """
        super(RecordList, self).__init__(parent, list_id)
        self._parent = parent
        # log.debug("RecordList %s: dir %s"%(list_id, self._entitydir))
        return

    def _migrate_filenames(self):
        """
        Override EntityData method
        """
        return None

    def _migrate_values(self, entitydata):
        """
        List description entity format migration method.

        The specification for this method is that it returns an entitydata value
        which is a copy of the supplied entitydata with format migrations applied.

        NOTE:  implementations are free to apply migrations in-place.  The resulting 
        entitydata should be exactly as the supplied data *should* appear in storage
        to conform to the current format of the data.  The migration function should 
        be idempotent; i.e.
            x._migrate_values(x._migrate_values(e)) == x._migrate_values(e)
        """
        for fkey, ftype in [(ANNAL.CURIE.display_type, "_enum_list_type")]:
            entitydata[fkey] = make_type_entity_id(
                ftype, extract_entity_id(entitydata[fkey])
                )
        migration_map = (
            [ (ANNAL.CURIE.record_type, ANNAL.CURIE.list_entity_type)
            ])
        entitydata = self._migrate_values_map_field_names(migration_map, entitydata)
        if ANNAL.CURIE.list_fields in entitydata:
            for f in entitydata[ANNAL.CURIE.list_fields]:
                field_id = extract_entity_id(f[ANNAL.CURIE.field_id])
                if field_id == "Field_render":
                    f[ANNAL.CURIE.field_id] = layout.FIELD_TYPEID+"/Field_render_type"
                if field_id == "Field_type":
                    f[ANNAL.CURIE.field_id] = layout.FIELD_TYPEID+"/Field_value_type"
                if field_id == "View_target_type":
                    f[ANNAL.CURIE.field_id] = layout.FIELD_TYPEID+"/View_entity_type"
                if field_id == "List_target_type":
                    f[ANNAL.CURIE.field_id] = layout.FIELD_TYPEID+"/List_entity_type"
        # Return result
        return entitydata

# End.



"""
Annalist collection

A collection is represented by:
- an ID (slug)
- a URI
- a name/label
- a description
- a set of record types
- a set of list views
- a set of record views
- ... and additional supporting metadata (fields, groups, user permissions, etc.)
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import os
import os.path
import shutil
import json
import datetime
from collections                    import OrderedDict
from distutils.version              import LooseVersion

from django.conf                    import settings

from utils.py3porting               import isoformat_space

import annalist
from annalist                               import layout
from annalist                               import message
from annalist.exceptions                    import Annalist_Error
from annalist.identifiers                   import RDF, RDFS, ANNAL
from annalist.util                          import valid_id, extract_entity_id, make_type_entity_id

from annalist.models.entity                 import Entity
from annalist.models.annalistuser           import AnnalistUser
from annalist.models.collectiontypecache    import CollectionTypeCache
from annalist.models.collectionfieldcache   import CollectionFieldCache
from annalist.models.collectionvocabcache   import CollectionVocabCache
from annalist.models.recordtype             import RecordType
from annalist.models.recordview             import RecordView
from annalist.models.recordlist             import RecordList
from annalist.models.recordfield            import RecordField
from annalist.models.recordgroup            import RecordGroup, RecordGroup_migration
from annalist.models.recordvocab            import RecordVocab
from annalist.models.rendertypeinfo         import (
    is_render_type_literal,
    is_render_type_id,
    is_render_type_set,
    is_render_type_list,
    is_render_type_object,
    )

#   ---------------------------------------------------------------------------
#
#   Static data for collection data caches
#
#   ---------------------------------------------------------------------------

type_cache  = CollectionTypeCache()
field_cache = CollectionFieldCache()
vocab_cache = CollectionVocabCache()

#   ---------------------------------------------------------------------------
#
#   Collection class
#
#   ---------------------------------------------------------------------------

class Collection(Entity):

    _entitytype     = ANNAL.CURIE.Collection
    _entitytypeid   = layout.COLL_TYPEID
    _entityview     = layout.SITE_COLL_VIEW
    _entityroot     = layout.SITE_COLL_PATH
    _entitybase     = layout.COLL_BASE_REF
    _entityfile     = layout.COLL_META_FILE
    _entityref      = layout.META_COLL_REF
    _contextbase    = layout.META_COLL_BASE_REF
    _contextref     = layout.COLL_CONTEXT_FILE

    def __init__(self, parentsite, coll_id, altparent=None):
        """
        Initialize a new Collection object.

        parentsite  is the parent site from which the new collection is descended.
        coll_id     the collection identifier for the collection
        altparent   is an alternative parent to search for descendents of the new 
                    Collection.  Effectively, the new Collection inherits definitions
                    from this alternative parent.
        """
        # log.debug("Collection.__init__: coll_id %s, parent dir %s"%(coll_id, parentsite._entitydir))
        if altparent is not None:
            if not isinstance(altparent, Collection):
                msg = "Collection altparent value must be a Collection (got %r)"%(altparent,)
                log.error(msg)
                raise ValueError(msg)
        self._parentsite = parentsite
        self._parentcoll = (
            altparent or
            None if coll_id == layout.SITEDATA_ID else 
            parentsite.site_data_collection()
            )
        super(Collection, self).__init__(parentsite, coll_id, altparent=self._parentcoll)
        return

    def _migrate_values(self, collmetadata):
        """
        Collection data format migration method.
        """
        migration_map = (
            [ (ANNAL.CURIE.comment,     ANNAL.CURIE.meta_comment    )
            ])
        collmetadata = self._migrate_values_map_field_names(migration_map, collmetadata)
        collmetadata[ANNAL.CURIE.id] = self._entityid # In case directory renamed by hand
        #@@ redundant?...
        if collmetadata[ANNAL.CURIE.type_id] == "_coll":
            collmetadata[ANNAL.CURIE.type_id] = self._entitytypeid
        #@@
        return collmetadata

    def flush_collection_caches(self):
        """
        Flush all caches associated with the current collection
        """
        type_cache.flush_cache(self)
        field_cache.flush_cache(self)
        vocab_cache.flush_cache(self)
        return

    @classmethod
    def flush_all_caches(self):
        """
        Flush all caches associated with all collections
        """
        type_cache.flush_all()
        field_cache.flush_all()
        vocab_cache.flush_all()
        return

    # Site

    def get_site(self):
        """
        Return site object for the site from which the current collection is accessed.
        """
        return self._parentsite

    def get_site_data(self):
        """
        Return parent object for accessing site data.
        """
        return self._parentsite.site_data_collection()

    # Alternate collections handling

    def set_alt_entities(self, altparent):
        """
        Update the alternative parent for the current collection.

        Returns a list of parents accessible from the supplied altparent (including itself)
        """
        # log.info("Collection.set_alt_entities: coll_id %s, altparent_id %s"%(self.get_id(), altparent.get_id()))
        if not isinstance(altparent, Collection):
            msg = "Collection.set_alt_entities value must be a Collection (got %r)"%(altparent,)
            log.error(msg)
            raise ValueError(msg)
        parents   = super(Collection, self).set_alt_entities(altparent)
        parentids = [ p.get_id() for p in parents ]
        # log.info(
        #     "@@ Collection.set_alt_entities: coll: %r, parentids %r"%
        #     (self.get_id(), parentids)
        #     )
        if layout.SITEDATA_ID not in parentids:
            msg = (
                "Entity.set_alt_entities cannot access site data (%s) via %r)"%
                (layout.SITEDATA_ID, altparent)
                )
            log.error(msg)
            raise ValueError(msg)
        if not self._ensure_values_loaded():
            msg = (
                "Entity.set_alt_entities cannot load collection data for %s)"%
                (self.get_id(),)
                )
            log.error(msg)
            raise ValueError(msg)
        self[ANNAL.CURIE.inherit_from] = make_type_entity_id(layout.COLL_TYPEID, altparent.get_id())
        if not (
            self.get(ANNAL.CURIE.default_list,        None) or 
            self.get(ANNAL.CURIE.default_view_type,   None) or 
            self.get(ANNAL.CURIE.default_view_entity, None)
            ):
            # Copy default collection view details from parent if none defined locally
            self[ANNAL.CURIE.default_list]        = altparent.get(ANNAL.CURIE.default_list,        None)
            self[ANNAL.CURIE.default_view_id]     = altparent.get(ANNAL.CURIE.default_view_id,     None)
            self[ANNAL.CURIE.default_view_type]   = altparent.get(ANNAL.CURIE.default_view_type,   None)
            self[ANNAL.CURIE.default_view_entity] = altparent.get(ANNAL.CURIE.default_view_entity, None)
        return parents

    @classmethod
    def create(cls, parent, coll_id, coll_meta):
        """
        Overload Entity.create with logic to set alternative parent details for 
        collection configuration inheritance, if an alternative is specified in 
        the collection data supplied.

        cls         is the Collection class object.
        parent      is the parent from which the collection is descended.
        coll_id     is the local identifier (slug) for the collection.
        coll_meta   is a dictionary of collection metadata values that are stored
                    for the created collection.

        Returns the created Collection instance.
        """
        # log.debug("Collection.create: %s, altscope %s"%(coll_id, altscope))
        coll = super(Collection, cls).create(parent, coll_id, coll_meta)
        if coll is not None:
            cls._set_alt_parent_coll(parent, coll)
        return coll

    @classmethod
    def _migrate_collection_config_dir(cls, parent, coll_id):
        # If old collection layout is present, migrate to new layout using "d/
        # Rename collection configuration directories and files individually so that 
        # existing data directories are not touched.
        parent_base_dir, parent_meta_file = parent._dir_path()
        coll_root_dir     = os.path.join(parent_base_dir, layout.SITE_COLL_PATH%{"id": coll_id})
        coll_base_dir     = os.path.join(coll_root_dir,   layout.COLL_BASE_DIR)
        coll_conf_old_dir = os.path.join(coll_root_dir,   layout.COLL_ROOT_CONF_OLD_DIR)
        #@@ TODO: remove this, not covered by tests.  Or remove entire method.
        if os.path.isdir(coll_conf_old_dir):
            log.info("Migrate old configuration from %s"%(coll_conf_old_dir,))
            for old_name in os.listdir(coll_conf_old_dir):
                old_path = os.path.join(coll_conf_old_dir, old_name)
                if ( ( os.path.isdir(old_path) ) or
                     ( os.path.isfile(old_path) and old_path.endswith(".jsonld") )
                   ):
                    log.info("- %s -> %s"%(old_name, coll_base_dir))
                    new_path = os.path.join(coll_base_dir, old_name)
                    try:
                        os.rename(old_path, new_path)
                    except Exception as e:
                        msg = (message.COLL_MIGRATE_DIR_FAILED%
                                { "id": coll_id
                                , "old_path": old_path, "new_path": new_path
                                , "exc": e
                                }
                            )
                        log.error("Collection._migrate_collection_config_dir: "+msg)
                        assert False, msg
            # Rename old config dir to avoid triggering this logic again
            coll_conf_saved_dir = coll_conf_old_dir+".saved"
            try:
                os.rename(coll_conf_old_dir, coll_conf_saved_dir)
            except Exception as e:
                msg = (message.COLL_MIGRATE_DIR_FAILED%
                        { "id": coll_id
                        , "old_path": coll_conf_old_dir, "new_path": coll_conf_saved_dir
                        , "exc": e
                        }
                    )
                log.error("Collection._migrate_collection_config_dir: "+msg)
                assert False, msg
        #@@
        return

    @classmethod
    def load(cls, parent, coll_id, altscope=None):
        """
        Overload Entity.load with logic to set alternative parent details for 
        collection configuration inheritance, if an alternative is specified in 
        the collection data loaded.

        cls         is the Collection class object.
        parent      is the parent from which the collection is descended.
        coll_id     is the local identifier (slug) for the collection.
        altscope    if supplied, indicates a scope other than the current collection
                    to search for children.

        Returns an instance of the indicated Collection class with data loaded from 
        the corresponding Annalist storage, or None if there is no such entity.
        """
        # log.debug("@@ Collection.load: %s, altscope %s"%(coll_id, altscope))
        cls._migrate_collection_config_dir(parent, coll_id)
        coll = super(Collection, cls).load(
            parent, coll_id, altscope=altscope
            )
        if coll is not None:
            cls._set_alt_parent_coll(parent, coll)
        return coll

    @classmethod
    def _set_alt_parent_coll(cls, parent, coll):
        """
        Set alternative parent collection - sets up search path for subsequent references.
        """
        coll_id        = coll.get_id()
        parent_coll_id = extract_entity_id(coll.get(ANNAL.CURIE.inherit_from, None))
        if parent_coll_id and parent_coll_id != layout.SITEDATA_ID:
            parent_coll = Collection.load(parent, parent_coll_id)
            if parent_coll is None:
                err_msg = message.COLL_PARENT_NOT_EXIST%{"id": coll_id, "parent_id": parent_coll_id}
                coll.set_error(err_msg)
                log.warning("Collection._set_alt_parent_coll: "+err_msg)
            else:
                log.debug(
                    "Collection._set_alt_parent_coll: coll %s references parent %s"%
                    (coll_id, parent_coll_id)
                    )
                coll.set_alt_entities(parent_coll)
        return coll

    # Software compatibility version

    def update_software_compatibility_version(self):
        # (assumes data loaded)
        ver = self.get(ANNAL.CURIE.software_version, None) or "0.0.0"
        if LooseVersion(ver) < LooseVersion(annalist.__version_data__):
            self[ANNAL.CURIE.software_version] = annalist.__version_data__
            self._save()

    # User permissions

    def create_user_permissions(self, user_id, user_uri,
            user_name, user_description,
            user_permissions=["VIEW"]
            ):
        user_values = (
            { ANNAL.CURIE.type:             ANNAL.CURIE.User
            , RDFS.CURIE.label:             user_name
            , RDFS.CURIE.comment:           user_description
            , ANNAL.CURIE.user_uri:         user_uri
            , ANNAL.CURIE.user_permission:  user_permissions
            })
        user = AnnalistUser.create(self, user_id, user_values)
        return user

    def get_user_permissions(self, user_id, user_uri):
        """
        Get a user permissions record (AnnalistUser).

        To return a value, both the user_id and the user_uri (typically a mailto: URI, but
        may be any *authenticated* identifier) must match.  This is to prevent access to 
        records of a deleted account being granted to a new account created with the 
        same user_id (username).

        user_id         local identifier for the type to retrieve.
        user_uri        authenticated identifier associated with the user_id.  That is,
                        the authentication service used is presumed to confirm that
                        the identifier belongs to the user currently logged in with
                        the supplied username.

        returns an AnnalistUser object for the identified user, or None.  This object contains
                information about permissions granted to the user in the current collection.
        """
        user = AnnalistUser.load(self, user_id, altscope="user")
        # log.debug("Collection.get_user_permissions: user_id %s, user_uri %s, user %r"%
        #     (user_id, user_uri, user)
        #     )
        if user:
            for f in [RDFS.CURIE.label, RDFS.CURIE.comment, ANNAL.CURIE.user_uri, ANNAL.CURIE.user_permission]:
                if f not in user:
                    user = None
                    break
        if user and user[ANNAL.CURIE.user_uri] != user_uri:
            user = None         # URI mismatch: return None.
        return user

    # Vocabulary namespaces

    def cache_get_vocab(self, vocab_id):
        """
        Retrieve namespace vocabulary entity for id (namespace prefix) from cache.

        Returns namespace vocabulary entity if found, otherwise None.
        """
        vocab_cache.get_vocab(self, vocab_id)
        t = vocab_cache.get_vocab(self, vocab_id)
        # Was it previously created but not cached?
        if not t and RecordType.exists(self, vocab_id, altscope="all"):
            msg = (
                "Collection.get_vocab %s present but not cached for collection %s"%
                (vocab_id, self.get_id())
                )
            log.warning(msg)
            t = RecordType.load(self, vocab_id, altscope="all")
            vocab_cache.set_vocab(self, t)
            # raise ValueError(msg) #@@@ (used in testing to help pinpoint errors)
        return t

    # Record types

    def types(self, altscope="all"):
        """
        Iterator over record types stored in the current collection.
        """
        return type_cache.get_all_types(self, altscope=altscope)

    def cache_add_type(self, type_entity):
        """
        Add or update type information in type cache.
        """
        log.debug("Collection.cache_add_type %s in %s"%(type_entity.get_id(), self.get_id()))
        type_cache.remove_type(self, type_entity.get_id())
        type_cache.set_type(self, type_entity)
        return

    def cache_get_type(self, type_id):
        """
        Retrieve type from cache.

        Returns type entity if found, otherwise None.
        """
        type_cache.get_type(self, type_id)
        t = type_cache.get_type(self, type_id)
        # Was it previously created but not cached?
        if not t and RecordType.exists(self, type_id, altscope="all"):
            msg = (
                "Collection.get_type %s present but not cached for collection %s"%
                (type_id, self.get_id())
                )
            log.warning(msg)
            t = RecordType.load(self, type_id, altscope="all")
            type_cache.set_type(self, t)
            # raise ValueError(msg) #@@@ (used in testing to help pinpoint errors)
        return t

    def cache_remove_type(self, type_id):
        """
        Remove type from type cache.
        """
        type_cache.remove_type(self, type_id)
        return

    def cache_get_all_type_ids(self, altscope="all"):
        """
        Iterator over type ids of types stored in the current collection.
        """
        return type_cache.get_all_type_ids(self, altscope=altscope)

    def cache_get_supertypes(self, type_entity):
        """
        Return supertypes of supplied type.
        """
        type_uri = type_entity.get_uri()
        return type_cache.get_type_uri_supertypes(self, type_uri)

    def cache_get_subtypes(self, type_entity):
        """
        Return subtypes of supplied type.
        """
        type_uri = type_entity.get_uri()
        return type_cache.get_type_uri_subtypes(self, type_uri)

    def cache_get_subtype_uris(self, type_uri):
        """
        Return subtype URIs of supplied type URI.

        The suplied URI is not itself required to be declared as identifying
        a defined type entity.
        """
        return type_cache.get_type_uri_subtype_uris(self, type_uri)

    def cache_get_supertype_uris(self, type_uri):
        """
        Return supertype URIs of supplied type URI.

        This returns all supertype URIs declared by the supertypes, even when 
        there is not corresponding type entity defined.
        """
        return type_cache.get_type_uri_supertype_uris(self, type_uri)

    def add_type(self, type_id, type_meta):
        """
        Add a new or updated record type to the current collection

        type_id     identifier for the new type, as a string
                    with a form that is valid as URI path segment.
        type_meta   a dictionary providing additional information about
                    the type to be created.

        Returns a RecordType object for the newly created type.
        """
        log.debug("Collection.add_type %s in %s"%(type_id, self.get_id()))
        t = RecordType.create(self, type_id, type_meta)
        return t

    def get_type(self, type_id):
        """
        Retrieve identified type description

        type_id     local identifier for the type to retrieve.

        returns a RecordType object for the identified type, or None.
        """
        if not valid_id(type_id):
            msg = "Collection %s get_type(%s) invalid id"%(self.get_id(), type_id)
            log.error(msg)
            raise ValueError(msg, type_id)
        return self.cache_get_type(type_id)

    def get_uri_type(self, type_uri):
        """
        Return type entity corresponding to the supplied type URI
        """
        t = type_cache.get_type_from_uri(self, type_uri)
        return t

    def remove_type(self, type_id):
        """
        Remove identified type description

        type_id     local identifier for the type to remove.

        Returns None on success, or a non-False status code if the type is not removed.
        """
        s = RecordType.remove(self, type_id)
        return s

    # def _unused_update_entity_types(self, e):
    #     # @@TODO: remove this?
    #     """
    #     Updates the list of type URIs associated with an entity by accessing the
    #     supertypes of the associated type record.
    #     """
    #     type_uri = e.get(ANNAL.CURIE.type, None)
    #     st_uris = [type_uri]
    #     t       = self.get_uri_type(type_uri)
    #     if t:
    #         assert (t.get_uri() == type_uri), "@@ type %s has unexpected URI"%(type_uri,)
    #         for st in type_cache.get_type_uri_supertypes(self, type_uri):
    #             st_uri = st.get_uri()
    #             if st_uri not in st_uris:
    #                 st_uris.append(st_uri)
    #     e['@type'] = st_uris
    #     return

    # Record views

    def views(self, altscope="all"):
        """
        Generator enumerates and returns record views that may be stored
        """
        for f in self._children(RecordView, altscope=altscope):
            v = self.get_view(f)
            if v and v.get_id() != layout.INITIAL_VALUES_ID:
                yield v
        return

    def add_view(self, view_id, view_meta):
        """
        Add a new record view to the current collection

        view_id     identifier for the new view, as a string
                    with a form that is valid as URI path segment.
        view_meta   a dictionary providing additional information about
                    the view to be created.

        returns a RecordView object for the newly created view.
        """
        v = RecordView.create(self, view_id, view_meta)
        return v

    def get_view(self, view_id):
        """
        Retrieve identified view description

        view_id     local identifier for the view to retrieve.

        returns a RecordView object for the identified view, or None.
        """
        v = RecordView.load(self, view_id, altscope="all")
        return v

    def remove_view(self, view_id):
        """
        Remove identified view description

        view_id     local identifier for the view to remove.

        Returns None on success, or a non-False status code if the view is not removed.
        """
        s = RecordView.remove(self, view_id)
        return s

    def set_default_view(self, view_id, type_id, entity_id):
        """
        Set and save the default list to be displayed for the current collection.
        """
        self[ANNAL.CURIE.default_view_id]     = view_id
        self[ANNAL.CURIE.default_view_type]   = type_id
        self[ANNAL.CURIE.default_view_entity] = entity_id
        self._save()
        return

    def get_default_view(self):
        """
        Return the default view id, type and entity to be displayed for the current collection.
        """
        view_id   = self.get(ANNAL.CURIE.default_view_id,     None)
        type_id   = self.get(ANNAL.CURIE.default_view_type,   None)
        entity_id = self.get(ANNAL.CURIE.default_view_entity, None)
        # log.info("Collection.get_default_view: %s/%s/%s"%(view_id, type_id, entity_id))
        return (view_id, type_id, entity_id) 

    # Record lists

    def lists(self, altscope="all"):
        """
        Generator enumerates and returns record lists that may be stored
        """
        for f in self._children(RecordList, altscope=altscope):
            l = self.get_list(f)
            if l and l.get_id() != layout.INITIAL_VALUES_ID:
                yield l
        return

    def add_list(self, list_id, list_meta):
        """
        Add a new record list to the current collection

        list_id     identifier for the new list, as a string
                    with a form that is valid as URI path segment.
        list_meta   a dictionary providing additional information about
                    the list to be created.

        returns a RecordList object for the newly created list.
        """
        l = RecordList.create(self, list_id, list_meta)
        return l

    def get_list(self, list_id):
        """
        Retrieve identified list description

        list_id     local identifier for the list to retrieve.

        returns a RecordList object for the identified list, or None.
        """
        l = RecordList.load(self, list_id, altscope="all")
        return l

    def remove_list(self, list_id):
        """
        Remove identified list description

        list_id     local identifier for the list to remove.

        Returns None on success, or a non-False status code if the list is not removed.
        """
        s = RecordList.remove(self, list_id)
        return s

    def set_default_list(self, list_id):
        """
        Set and save the default list to be displayed for the current collection.
        """
        self[ANNAL.CURIE.default_list]        = list_id
        self[ANNAL.CURIE.default_view_id]     = None
        self[ANNAL.CURIE.default_view_type]   = None
        self[ANNAL.CURIE.default_view_entity] = None
        self._save()
        return

    def get_default_list(self):
        """
        Return the default list to be displayed for the current collection.
        """
        list_id = self.get(ANNAL.CURIE.default_list, None)
        if list_id and not RecordList.exists(self, list_id, altscope="all"):
            log.warning(
                "Default list %s for collection %s does not exist"%
                (list_id, self.get_id())
                )
            list_id = None
        return list_id 

    # View (and list) fields and properties

    #@@
    # @classmethod
    # def reset_field_cache(cls):
    #     """
    #     Used for testing: clear out field cache so tets don't interfere with each 
    #     other.  Could also be used when external application updates data.
    #     """
    #     field_cache.flush_all()
    #     return
    #@@

    def fields(self, altscope="all"):
        """
        Iterator over view fields stored in the current collection.
        """
        return field_cache.get_all_fields(self, altscope=altscope)

    def cache_add_field(self, field_entity):
        """
        Add or update field information in field cache.
        """
        log.debug("Collection.cache_add_field %s in %s"%(field_entity.get_id(), self.get_id()))
        field_cache.remove_field(self, field_entity.get_id())
        field_cache.set_field(self, field_entity)
        return

    def cache_get_field(self, field_id):
        """
        Retrieve field from cache.

        Returns field entity if found, otherwise None.
        """
        #@@ field_cache.get_field(self, field_id)
        t = field_cache.get_field(self, field_id)
        # Was it previously created but not cached?
        if not t and RecordField.exists(self, field_id, altscope="all"):
            msg = (
                "Collection.get_field %s present but not cached for collection %s"%
                (field_id, self.get_id())
                )
            log.warning(msg)
            t = RecordField.load(self, field_id, altscope="all")
            field_cache.set_field(self, t)
        return t

    def cache_remove_field(self, field_id):
        """
        Remove field from field cache.
        """
        field_cache.remove_field(self, field_id)
        return

    def cache_get_all_field_ids(self, altscope="all"):
        """
        Iterator over field ids of fields stored in the current collection.
        """
        return field_cache.get_all_field_ids(self, altscope=altscope)

    def cache_get_subproperty_fields(self, field_entity):
        """
        Return fields that use subproperties of supplied field's property URI.
        """
        property_uri = field_entity.get_property_uri()
        return field_cache.get_subproperty_fields(self, property_uri)

    def cache_get_superproperty_fields(self, field_entity):
        """
        Return fields that use superproperties of supplied field's property URI.
        """
        property_uri = field_entity.get_property_uri()
        return field_cache.get_superproperty_fields(self, property_uri)

    def cache_get_subproperty_uris(self, property_uri):
        """
        Return subproperty URIs of supplied property URI.

        The suplied URI is not itself required to be declared as used by
        a defined field entity.
        """
        return field_cache.get_subproperty_uris(self, property_uri)

    def cache_get_superproperty_uris(self, property_uri):
        """
        Return superproperty URIs of supplied property URI.

        This returns all superproperty URIs declared by the field, and any
        fields that use the superproperty URIs, even when there is no 
        corresponding field definition.
        """
        return field_cache.get_superproperty_uris(self, property_uri)

    def add_field(self, field_id, field_meta):
        """
        Add a new or updated record field to the current collection

        field_id    identifier for the new field, as a string
                    with a form that is valid as URI path segment.
        field_meta  a dictionary providing additional information about
                    the field to be created.

        Returns a RecordField object for the newly created field.
        """
        log.debug("Collection.add_field %s in %s"%(field_id, self.get_id()))
        f = RecordField.create(self, field_id, field_meta)
        return f

    def get_field(self, field_id):
        """
        Retrieve identified field description

        field_id    local identifier for the field to retrieve.

        returns a RecordField object for the identified field, or None.
        """
        if not valid_id(field_id, reserved_ok=True):
            msg = "Collection %s get_field(%s) invalid id"%(self.get_id(), field_id)
            log.error(msg)
            # Construct and return a placeholder field
            ph_meta = (
                { RDFS.CURIE.label:               "(field error)"
                , ANNAL.CURIE.field_render_type:  "_enum_render_type/Placeholder"
                , ANNAL.CURIE.field_value_mode:   "_enum_value_mode/Value_direct"
                , ANNAL.CURIE.field_placement:    "small:0,12"
                , ANNAL.CURIE.placeholder:        
                    "(Invalid field id: '%s')"%(field_id,)
                })
            f = RecordField._child_init(self, "_placeholder")
            f.set_values(ph_meta)
            return f
        return self.cache_get_field(field_id)

    def get_uri_field(self, property_uri):
        """
        Return field entity corresponding to the supplied property URI
        """
        t = field_cache.get_field_from_property_uri(self, property_uri)
        return t

    def remove_field(self, field_id):
        """
        Remove identified field description

        field_id     local identifier for the field to remove.

        Returns None on success, or a non-False status code if the field is not removed.
        """
        s = RecordField.remove(self, field_id)
        return s

    # JSON-LD context data

    def generate_coll_jsonld_context(self, flags=None):
        """
        (Re)generate JSON-LD context description for the current collection.

        Returns list of errors, or empty list.
        """
        errs = []
        if flags and ("nocontext" in flags):
            # Skip processing if "nocontext" flag provided
            return
        # log.info("Generating context for collection %s"%(self.get_id()))
        # Build context data
        context      = self.get_coll_jsonld_context()
        datetime_now = datetime.datetime.today().replace(microsecond=0)
        datetime_str = isoformat_space(datetime_now)
        # Assemble and write out context description
        with self._metaobj(
                layout.META_COLL_BASE_REF,
                layout.COLL_CONTEXT_FILE,
                "wt"
                ) as context_io:
            json.dump(
                { "_comment": "Generated by generate_coll_jsonld_context on %s"%datetime_str
                , "@context": context 
                }, 
                context_io, indent=2, separators=(',', ': '), sort_keys=True
                )
        # Create collection README.md for human context...
        if self._values:
            README_vals = (
                { "id":         self.get_id()
                , "label":      self._values.get("rdfs:label", self.get_id())
                , "heading":    ""
                , "comment":    self._values.get("rdfs:comment", "")
                })
            if not README_vals["comment"].startswith("#"):
                README_vals["heading"] = message.COLL_README_HEAD%README_vals
            README_text = message.COLL_README%README_vals
            with self._metaobj(
                    layout.META_COLL_REF,
                    "README.md",
                    "wt"
                    ) as readme_io:
                readme_io.write(README_text)
        return errs

    def get_coll_jsonld_context(self):
        """
        Return dictionary containing context structure for collection.

        Entry '@errs' is set to a list of errors encountered, or an empty list.
        """
        # Use OrderedDict to allow some control over ordering of context file contents:
        # this is for humane purposes only, and is not technically critical.
        errs              = []
        context           = OrderedDict(
            # { "@base":                  self.get_url() + layout.META_COLL_BASE_REF
            { ANNAL.CURIE.type:         { "@type":      "@id"   }
            , ANNAL.CURIE.entity_list:  { "@container": "@list" }
            })
        # Collection-local URI prefix
        context.update(
            { '_site_':         self.get_site().get_url()
            , '_coll_':         self.get_url()
            , '_base_':         self.get_url() + layout.COLL_BASE_REF
            })
        # Common import/upload fields
        context.update(
            { 'resource_name':  "annal:resource_name"
            , 'resource_type':  "annal:resource_type"
            })
        # upload-file fields
        context.update(
            { 'upload_name':    "annal:upload_name"
            , 'uploaded_file':  "annal:uploaded_file"
            , 'uploaded_size':  "annal:uploaded_size"
            })
        # import-resource fields
        context.update(
            { 'import_name':    "annal:import_name"
            , 'import_url':    
              { "@id":          "annal:import_url"
              , "@type":        "@id"
              }
            })
        # Scan types, generate prefix data
        for t in self.child_entities(RecordType, altscope="all"):
            tid = t.get_id()
            if tid != layout.INITIAL_VALUES_ID:
                tns = t.get(ANNAL.CURIE.ns_prefix, "")
                if tns != "":
                    context[tns] = self.get_url() + layout.COLL_TYPEDATA_VIEW%({"id": tid})
        # Scan vocabs, generate prefix data (possibly overriding type-derived data)
        for v in self.child_entities(RecordVocab, altscope="all"):
            vid = v.get_id()
            if vid != layout.INITIAL_VALUES_ID:
                if ANNAL.CURIE.uri in v:
                    vuri = v[ANNAL.CURIE.uri]
                    if vuri[-1] not in {":", "/", "?", "#"}:
                        msg  = (
                            "Vocabulary %s namespace URI %s does not end with an expected delimiter"%
                            (vid, vuri)
                            )
                        log.warning(msg)
                        errs.append(msg)
                    context[vid] = v[ANNAL.CURIE.uri]
        # Scan view fields and generate context data for property URIs used
        for v in self.child_entities(RecordView, altscope="all"):
            view_fields = v.get(ANNAL.CURIE.view_fields, [])
            for fref in view_fields:
                fid  = extract_entity_id(fref[ANNAL.CURIE.field_id])
                vuri = fref.get(ANNAL.CURIE.property_uri, None)
                furi, fcontext, field_list = self.get_field_uri_jsonld_context(
                    fid, self.get_field_jsonld_context
                    )
                if fcontext is not None:
                    fcontext['vid'] = v.get_id()
                    fcontext['fid'] = fid
                e = self.set_field_uri_jsonld_context(vuri or furi, fid, fcontext, context)
                errs.extend(e)
                # If this field contains a list of subfields, scan those
                # NOTE: current implementation handles only a single level of field nesting
                if field_list:
                    for subfref in field_list:
                        subfid  = extract_entity_id(subfref[ANNAL.CURIE.field_id])
                        subfuri = subfref.get(ANNAL.CURIE.property_uri, None)
                        furi, fcontext, field_list = self.get_field_uri_jsonld_context(
                            subfid, self.get_field_jsonld_context
                            )
                        if fcontext is not None:
                            fcontext['fid']    = fid
                            fcontext['subfid'] = subfid
                        e = self.set_field_uri_jsonld_context(subfuri or furi, subfid, fcontext, context)
                        errs.extend(e)
        # Scan group fields and generate context data for property URIs used
        #@@TODO - to be deprecated
        # In due course, field groups will replaced by inline field lists.
        # This code does not process field lists for fields referenced by a group.
        #@@
        for g in self.child_entities(RecordGroup_migration, altscope="all"):
            for gref in g[ANNAL.CURIE.group_fields]:
                fid  = extract_entity_id(gref[ANNAL.CURIE.field_id])
                guri = gref.get(ANNAL.CURIE.property_uri, None)
                furi, fcontext, field_list = self.get_field_uri_jsonld_context(
                    fid, self.get_field_jsonld_context
                    )
                if fcontext is not None:
                    fcontext['gid'] = g.get_id()
                    fcontext['fid'] = fid
                e = self.set_field_uri_jsonld_context(guri or furi, fid, fcontext, context)
                errs.extend(e)
        if errs:
            context['@errs'] = errs
        return context

    def get_field_uri_jsonld_context(self, fid, get_field_context):
        """
        Access field description, and return field property URI and appropriate 
        property description for JSON-LD context.

        Returns a triple consisting of the field property URI, the context information to be
        generated for the field, and a list of any field references contained directly within
        the field definition (as opposed to a field group reference)

        If there is no corresponding field description, returns (None, None, None)

        If no context should be generated for the field URI, returns (uri, None, field_list)

        The field list returned is 'None' if there is no contained list of fields.
        """
        f = RecordField.load(self, fid, altscope="all")
        if f is None:
            return (None, None, None)
        field_list = f.get(ANNAL.CURIE.field_fields, None)
        return (f[ANNAL.CURIE.property_uri], get_field_context(f), field_list)

    def set_field_uri_jsonld_context(self, puri, field_id, fcontext, property_contexts):
        """
        Save property context description into supplied property_contexts dictionary.  
        If the context is already defined, generate warning if there is a compatibility 
        problem.

        Returns list of errors, or empty list.
        """
        errs = []
        if puri:
            uri_parts = puri.split(":")
            if len(uri_parts) > 1:    # Ignore URIs without ':'
                if not fcontext:
                    # For diagnostics to locate incompatible use...
                    fcontext = {'fid': field_id}
                if (puri in property_contexts):
                    pcontext    = property_contexts[puri]
                    pcontext.pop('err', None)   # Drop pevious error(s) from report
                    p_type      = pcontext.get("@type", None)
                    f_type      = fcontext.get("@type", None)
                    if (p_type != f_type):
                        msg  = (
                            "Incompatible value type for property %s in field %s (new %r; was %r)"%
                            (puri, field_id, fcontext, pcontext)
                            )
                        log.warning(msg)
                        property_contexts[puri]['err'] = msg
                        errs.append(msg)
                    p_container = pcontext.get("@container", None)
                    f_container = fcontext.get("@container", None)
                    if ( (p_container != f_container) and
                         ( (p_container == "@list") or 
                           (f_container == "@list") ) ):
                        msg  = (
                            "Incompatible container type for property %s in field %s (new %r; was %r)"%
                            (puri, field_id, fcontext, pcontext)
                            )
                        # msgp = "pcontext @type %s, @container %s"%(p_type, p_container)
                        # msgf = "fcontext @type %s, @container %s"%(f_type, f_container)
                        log.warning(msg)
                        # print "@@ "+msg
                        # print "@@ pcontext @type %s, @container %s"%(p_type, p_container)
                        # print "@@ fcontext @type %s, @container %s"%(f_type, f_container)
                        property_contexts[puri]['err'] = msg
                        errs.append(msg)
                        # errs.append(msgp)
                        # errs.append(msgf)
                elif ( fcontext and
                       ( uri_parts[0] in property_contexts ) or         # Prefix defined vocab?
                       ( uri_parts[0] in ["http", "https", "file"] ) ): # Full URI?
                    property_contexts[puri] = fcontext
                    # msg = "Save context info for %s in field %s (new %r)"% (puri, field_id, fcontext)
                    # print "@@ "+msg
        return errs

    # @@TODO: move this away from model logic, as it represents a dependency on view logic?
    @staticmethod
    def get_field_jsonld_context(fdesc):
        """
        Returns a context description for the supplied field description.

        Returns None if no property context information is needed for the 
        supplied field.
        """
        rtype = extract_entity_id(fdesc[ANNAL.CURIE.field_render_type])
        vmode = extract_entity_id(fdesc[ANNAL.CURIE.field_value_mode])
        if vmode in ["Value_entity", "Value_field"]:
            rtype = "Enum"
        elif vmode == "Value_import":
            rtype = "URIImport"
        elif vmode == "Value_upload":
            rtype = "FileUpload"

        if is_render_type_literal(rtype):
            fcontext = {} # { "@type": "xsd:string" }
        elif is_render_type_id(rtype):
            fcontext = { "@type": "@id" }   # Add type from field descr?
        elif is_render_type_object(rtype):
            fcontext = {}
        else:
            msg = "Unexpected value mode or render type (%s, %s)"%(vmode, rtype)
            log.error(msg)
            raise ValueError(msg)

        if is_render_type_set(rtype):
            fcontext["@container"] = "@set"
        elif is_render_type_list(rtype):
            fcontext["@container"] = "@list"

        return fcontext

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Renderer and value mapper for Entity Id field
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2016, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

from django.template        import Template

from utils.py3porting                   import is_string, to_unicode

from annalist.views.fields.render_base  import RenderBase

#   ----------------------------------------------------------------------------
#
#   Entity Id value mapping
#
#   ----------------------------------------------------------------------------

class EntityIdValueMapper(RenderBase):
    """
    Value mapper class for entity id field.
    """

    @classmethod
    def decode(cls, field_value):
        """
        Returns an entity Id form field value as itself with leading/trailing spaces trimmed
        """
        if is_string(field_value):
            field_value = field_value.strip()
        return field_value

    # encode defaults to identity mapper]

#   ----------------------------------------------------------------------------
#
#   Entity Id field renderers
#
#   ----------------------------------------------------------------------------

# End.



"""
Personal deployment (same host) settings

Data is kept in personal home directory area.
Service configuration is kept under personal home directory.
This is also useful for non-system deployment on a shared host.
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

from .common import *

ANNALIST_VERSION_MSG = "Annalist version %s (personal_debug configuration)"%(ANNALIST_VERSION)

SETTINGS_MODULE   = __name__
SITE_DIR_NAME     = "annalist_site"
BASE_DATA_DIR     = os.path.expanduser("~")
BASE_SITE_DIR     = os.path.join(BASE_DATA_DIR, SITE_DIR_NAME)
CONFIG_BASE       = os.path.join(os.path.expanduser("~"), ".annalist/")
STATIC_ROOT       = os.path.join(BASE_SITE_DIR, 'static')

BASE_LOG_DIR      = BASE_SITE_DIR+"/"
ANNALIST_LOG_PATH = BASE_LOG_DIR+ANNALIST_LOG_FILE
ACCESS_LOG_PATH   = BASE_LOG_DIR+ACCESS_LOG_FILE
ERROR_LOG_PATH    = BASE_LOG_DIR+ERROR_LOG_FILE

DATABASE_PATH     = os.path.join(BASE_SITE_DIR, 'db.sqlite3')
DATABASES         = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': DATABASE_PATH,
    }
}

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG           = True

ALLOWED_HOSTS   = ['*']     # Insecure: use e.g. ['.annalist.net']

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '%(levelname)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s'
        },
        'timed': {
            'format': '%(levelname)s %(asctime)s %(message)s'
        },
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
    },
    'handlers': {
        # Include the default Django email handler for errors
        # This is what you'd get without configuring logging at all.
        'mail_admins': {
            'class': 'django.utils.log.AdminEmailHandler',
            'level': 'ERROR',
             # But the emails are plain text by default - HTML is nicer
            'include_html': True,
            'formatter': 'verbose'
        },
        # Log to a text file that can be rotated by logrotate
        'logfile': {
            # 'class': 'logging.handlers.WatchedFileHandler',
            # 'class':        'logging.handlers.RotatingFileHandler',
            'class':        'annalist_site.settings.common.RotatingNewFileHandler',
            'filename':     ANNALIST_LOG_PATH,
            'maxBytes':     2*1024*1024,            # 2Mb
            'backupCount':  9,                      # Keep 9 files
            'level':        'DEBUG',
            'formatter':    'timed'
        },
    },
    'loggers': {
        # Again, default Django configuration to email unhandled exceptions
        # 'django.request': {
        #     'handlers': ['mail_admins'],
        #     'level': 'ERROR',
        #     'propagate': True,
        # },
        'django.request': {
            'handlers': ['logfile'],
            'level': 'INFO',
            'propagate': True,
        },
        # Might as well log any errors anywhere else in Django
        'django': {
            'handlers': ['logfile'],
            'level': 'INFO',
            'propagate': False,
        },
        'annalist_root': {
            'handlers': ['logfile'],
            'level': 'INFO',
            'propagate': False
        },
        'annalist_site': {
            'handlers': ['logfile'],
            'level': 'INFO',
            'propagate': False
        },
        'annalist': {
            'handlers': ['logfile'],
            'level': 'DEBUG',
            'propagate': False
        },
        'login': {
            'handlers': ['logfile'],
            'level': 'DEBUG',
            'propagate': False
        },
    },
}

import logging
log = logging.getLogger(__name__)
log.info("Annalist starting...")

# Force new log files for any rotating file log handlers
for h in log.handlers:
    log.info("@@ log handler %r"%(h,))
    if isinstance(h, logging.handlers.RotatingFileHandler):
        log.info("@@ log rollover")        
        h.doRollover()

# log.info("Annalist version %s (personal configuration)"%(ANNALIST_VERSION))
log.info(ANNALIST_VERSION_MSG)
log.debug("SETTINGS_MODULE:   "+SETTINGS_MODULE)
log.debug("BASE_DATA_DIR:     "+BASE_DATA_DIR)
log.debug("CONFIG_BASE:       "+CONFIG_BASE)
log.debug("DJANGO_ROOT:       "+DJANGO_ROOT)
log.debug("SITE_CONFIG_DIR:   "+SITE_CONFIG_DIR)
log.debug("SITE_SRC_ROOT:     "+SITE_SRC_ROOT)
log.debug("STATICFILES_DIRS:  "+repr(STATICFILES_DIRS))
log.debug("DB PATH:           "+DATABASES['default']['NAME'])
log.debug("ALLOWED_HOSTS:     "+",".join(ALLOWED_HOSTS))
log.debug("ANNALIST_LOG_PATH: "+ANNALIST_LOG_PATH)
log.debug("TRACE_FIELD_VALUE: "+str(TRACE_FIELD_VALUE))
# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Annalist action confirmation view definition
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

# import os.path
# import json
# import random
import logging
# import uuid
# import copy

import logging
log = logging.getLogger(__name__)

# import rdflib
# import httplib2

from django.http                    import HttpResponse
from django.http                    import HttpResponseRedirect
from django.template                import loader

from django.conf                    import settings

from utils.ContentNegotiationView   import ContentNegotiationView

from annalist.models.annalistuser   import AnnalistUser
from annalist.models.annalistuser   import (
    site_default_user_id, site_default_user_uri, 
    default_user_id, default_user_uri, 
    unknown_user_id, unknown_user_uri
    )

from annalist.views.generic         import AnnalistGenericView

class ProfileView(AnnalistGenericView):
    """
    View class to handle requests to the Annalist user profile URI
    """
    def __init__(self):
        super(ProfileView, self).__init__()
        return

    # GET

    def get(self, request):
        def resultdata():
            username, useruri = self.get_user_identity()
            recent_userid = request.session.get('recent_userid', username)
            return (
                { 'title':              self.site_data()["title"]
                , 'user':               request.user 
                , 'username':           username
                , 'useruri':            useruri
                , 'userid':             recent_userid
                , 'continuation_url':   continuation_url
                })
        continuation_url  = self.continuation_next(
            request.GET, self.view_uri("AnnalistHomeView")
            )
        return (
            self.authenticate(continuation_url) or 
            self.render_html(resultdata(), 'annalist_profile.html') or 
            self.error(self.error406values())
            )

    def post(self, request):
        if request.POST.get('continue', None):
            # Check if user permissions are defined
            user_id, user_uri = self.get_user_identity()
            site_coll  = self.site().site_data_collection()
            user_perms = site_coll.get_user_permissions(user_id, user_uri)
            if not user_perms:
                # Initialize new user permissions from system defaults
                # (NOTE: site-specific default permissions are incorporate dynamically)
                default_perms = site_coll.get_user_permissions(
                    default_user_id, default_user_uri
                    )
                new_perms_values = default_perms.get_values()
                new_perms_values.update(
                    { "annal:id":               None
                    , "rdfs:label":             "Permissions for %s"%user_id
                    , "rdfs:comment":           "# Permissions for %s\r\n\r\n"%user_id+
                                                "Permissions for user %s (copied from default)"%user_id
                    , "annal:user_uri":         user_uri
                    })
                new_perms = AnnalistUser.create(site_coll, user_id, new_perms_values)
        continuation_url  = request.POST.get("continuation_url", "../")
        return HttpResponseRedirect(continuation_url)

# End.



from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
Annalist action confirmation view definition
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import os
import logging

from django.http                    import HttpResponse
from django.http                    import HttpResponseRedirect
from django.template                import loader

from django.conf                    import settings

from utils.ContentNegotiationView   import ContentNegotiationView

from annalist.models.annalistuser   import AnnalistUser

from annalist.views.generic         import AnnalistGenericView

# Helper function to get last N lines of log file, rather than loading
# entire file into memory (was getting MemoryError falures.)
#
# Adapted from: https://stackoverflow.com/a/13790289/324122

def tail(f, lines=1, _buffer=4098):
    """
    Tail a file and get X lines from the end
    """
    lines_found = []
    # block counter will be multiplied by buffer
    # to get the block size from the end
    block_counter = -1
    # loop until we find more than X lines
    # (Looking for `>lines` because the first line may be truncated)
    while len(lines_found) <= lines:
        try:
            f.seek(block_counter * _buffer, os.SEEK_END)
        except IOError:  # either file is too small, or too many lines requested
            f.seek(0)
            lines_found = f.readlines()
            break
        lines_found = f.readlines()
        block_counter -= 1
    return lines_found[-lines:]


class ServerLogView(AnnalistGenericView):
    """
    View class to handle requests to the Annalist user profile URI
    """
    def __init__(self):
        super(ServerLogView, self).__init__()
        return

    # GET

    def get(self, request):
        def resultdata():
            serverlogname = settings.ANNALIST_LOG_PATH
            log.info("ServerLogView: serverlogname %s"%(serverlogname,))
            with open(serverlogname, "r") as serverlogfile:
                # serverlog     = list(serverlogfile) # Generates MemoryError with large logs
                serverlog = tail(serverlogfile, 2000)
                return (
                    { 'title':              self.site_data()["title"]
                    , 'serverlogname':      serverlogname
                    , 'serverlog':          "".join(serverlog)
                    , 'continuation_url':   continuation_url
                    })

        continuation_url  = self.continuation_next(
            request.GET, self.view_uri("AnnalistHomeView")
            )
        return (
            # self.authenticate(continuation_url) or 
            self.authorize("ADMIN", None) or
            self.render_html(resultdata(), 'annalist_serverlog.html') or 
            self.error(self.error406values())
            )

    def post(self, request):
        continuation_url  = request.POST.get("continuation_url", "../")
        return HttpResponseRedirect(continuation_url)

# End.



"""
Authentication backend using Credential object returned by oauth2client flow exchange
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import json
import copy
import re

import logging
log = logging.getLogger(__name__)

import httplib2

from django.contrib.auth.models import User

from utils.py3porting import is_string, to_unicode

class OAuth2CheckBackend(object):
    """
    Authenticate using credential object from OAuth2 exchange

    username is a local user id that keys the local user database
    password is a Credential object obtained via the OAuth2 dance
    profile  is a user profile information dictionary

    NOTE: when this method returns a User record on completion of a third
    party authentication process, it does not guarantee that it is the same
    as any record that may have been previously associated with the supplied
    username.  It becomes the responsibility of the calling view code to check 
    that the user details match any previously associated with the user id.

    The returned user object is created or copied from the Django user base,
    but if it already exists the email address is replaced with the one
    returned by the OIDC authentication exchange.
    """

    def authenticate(self, username=None, profile=None):
        log.debug(
            "OAuth2CheckBackend.authenticate: username %s, profile %r"%
            (username, profile)
            )
        if is_string(profile):
            # Not oauth2 exchange:
            # @TODO: can we be more specific about what type this should be?
            return None
        auth_username   = None
        auth_email      = None
        auth_url        = None
        return_user     = None
        create_username = None
        verified_email  = False
        verified_url    = False
        if profile:
            # Analyze profile returned for verified identifier
            # (email address or URL)
            #
            # Google returns a "verified_email" flag
            # (It looks like this was changed in Google profile...)
            log.info("login user profile: %r"%(profile,))
            verified_email = (
                profile.get("verified_email", False) or 
                profile.get("email_verified", False)
                )
            if verified_email:
                # Use access token to retrieve profile information
                # Construct authenticated user ID from email local part
                auth_email       = profile["email"]
                email_local_part = auth_email.split('@', 1)[0]
                auth_username    = re.sub(r"\.", "_", email_local_part)
                auth_username    = re.sub(r"[^a-zA-Z0-9_]", "", auth_username)
                auth_username    = auth_username[:32]
            # GitHub returns a URL that is implicitly verified
            # NOTE: the GitHub email does not include a separate flag to say
            #       it's verified, though apparently it is.  But the GitHub
            #       email address is optional in the profile.
            if profile.get("html_url", "").startswith("https://github.com/"):
                verified_url  = True    
                auth_url      = profile["html_url"]
                auth_domain   = "github.com"
                auth_username = profile["login"]
                # This is a hack to fit Django's User structure:
                # would prefer to store and use authenticated URL instead.
                # Maybe this is possible?
                # See https://stackoverflow.com/questions/6085025/django-user-profile
                auth_email    = (
                    profile.get("email", None) or
                    auth_username+"."+auth_domain+"@user.annalist.net" 
                    )
        if username:
            try:
                return_user       = User.objects.get(username=username)
                return_user.email = auth_email
            except User.DoesNotExist:
                create_username = username
        elif auth_username:
            try:
                return_user = User.objects.get(username=auth_username)
            except User.DoesNotExist:
                create_username = auth_username
        if create_username:
            # NOTE: when a new User record is created, it is important
            # that it is saved to the local Django database before
            # returning it to the caller, to ensure that a suitable
            # primary key value is created.  The authentication return
            # path makes further changes to the User record which cause
            # the Django ORM to force an update rather than insert of the
            # new record, which in turn generates an error if no primary
            # key is defined.
            log.info("Create new user record for %s"%(create_username))
            return_user = User(username=create_username, password='Not specified')
            if profile is not None:
                return_user.is_staff     = True
                return_user.is_superuser = False
                #@@ For testing: fake old-style Google profile
                # if ("given_name" in profile) and ("family_name" in profile):
                #     profile["name"] = profile["given_name"] + " " + profile["family_name"]
                #     del profile["given_name"]
                #     del profile["family_name"]
                #@@
                if ("given_name" in profile) and ("family_name" in profile):
                    given_name  = profile["given_name"]
                    family_name = profile["family_name"]
                elif ("name" in profile) and profile["name"]:
                    # Older Google profiles have just "name" value, apparently
                    # GitHub profiles may include a "name" value, a[parently]
                    n = profile["name"].split(None, 1)
                    given_name  = ""
                    family_name = ""
                    if len(n) >= 1:
                        given_name  = n[0]
                    if len(n) >= 2:
                        family_name = n[1]
                else:
                    given_name  = auth_username
                    family_name = ""
                return_user.first_name   = given_name
                return_user.last_name    = family_name
                return_user.email        = auth_email
            elif password.id_token:
                # No profile provided: Try to load email address from id_token
                return_user.is_staff     = True
                return_user.is_superuser = False
                return_user.first_name   = ""
                return_user.last_name    = ""
                return_user.email        = password.id_token["email"]
            else:
                return_user = None
            if return_user:
                return_user.save()
        if return_user:
            log.info("user.username:   %s"%(return_user.username,))
            log.info("user.first_name: %s"%(return_user.first_name,))
            log.info("user.last_name:  %s"%(return_user.last_name,))
            log.info("user.email:      %s"%(return_user.email,))
            return return_user
        # No username or credentials provided
        log.info("No user id or no credentials provided")
        return None

    def get_user(self, user_id):
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None

# End.



"""
Django settings for annalist_site project.

For more information on this file, see
https://docs.djangoproject.com/en/1.6/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.6/ref/settings/
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2011-2013, University of Oxford"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import os
import django
import sys
import string
import random
import logging
import logging.handlers
from annalist import __version__
from annalist import layout

DJANGO_ROOT     = os.path.dirname(os.path.realpath(django.__file__))
SETTINGS_DIR    = os.path.dirname(os.path.realpath(__file__))           # src/annalist_root/annalist_site/settings
SITE_CONFIG_DIR = os.path.dirname(SETTINGS_DIR)                         # src/annalist_root/annalist_site
SITE_SRC_ROOT   = os.path.dirname(SITE_CONFIG_DIR)                      # src/annalist_root
SAMPLEDATA_DIR  = SITE_SRC_ROOT+"/sampledata/data"                      # src/annalist_root/sampledata

ANNALIST_SITE_SEG = "annalist"              # Base URL path segment for Annalist site
ANNALIST_SITE_REF = ANNALIST_SITE_SEG+"/"   # Base URL path for Annalist site
ANNALIST_LOG_FILE = "annalist.log"
ACCESS_LOG_FILE   = "annalist-wsgi-access.log"
ERROR_LOG_FILE    = "annalist-wsgi-error.log"

class RotatingNewFileHandler(logging.handlers.RotatingFileHandler):
    """
    Define a rotating file logging handler that additionally forces a new file 
    the first time it is instantiated in a run of the containing program.

    NOTE: if multiple file handlers are used with in an application, only the
    first one instantiated will be allocated a new file at startup.  The
    class variable '_newfile' might be replaced with a dictionary
    indexed by the (fully expanded) filename.
    """

    _newfile = False

    def __init__(self, *args, **kwargs):
        super(RotatingNewFileHandler, self).__init__(*args, **kwargs)
        if not RotatingNewFileHandler._newfile:
            self.doRollover()
            RotatingNewFileHandler._newfile = True
        return

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.6/howto/deployment/checklist/

# Define secret key for various Django functions
# See: 
#   https://stackoverflow.com/a/15383766/324122
#   https://stackoverflow.com/a/23728630/324122
#
# SECURITY WARNING: keep the secret key used in production secret!
KEY_CHARS  = string.ascii_letters + string.digits + string.punctuation
SECRET_KEY = ''.join(
                random.SystemRandom().choice(KEY_CHARS) 
                for _ in range(32)
                )
# See also: https://stackoverflow.com/a/23728630/324122
# SECRET_KEY = '@-+h*%@h+0yj(^c9y-=1a@9l^@xzub200ofq2@a$gm2k_l*$pf'

# UPDATE: running under 'gunicorn', I've found that session logins get dumped
# periodcally when the worker process periodically restarts.  
# Hence I'm trying to find an alternative that attempts to be more "sticky", 
# without depending on a predefinbed secret key.
#
# Options I'm considering:
# (a) use an environment variable set randomly when starting the server,
# (b) use the file system to persist a randomly generated value
#
# Currently, I prefer the environment variable approach:
# For this to work, set the environment variable in 'annalist-manager' when 
# running under 'gunicorn'. If not defined, use randomly key (from above).

if "ANNALIST_KEY" in os.environ:
    SECRET_KEY = os.environ["ANNALIST_KEY"]

# SECURITY WARNING: don't run with debug turned on in production!
# (overrides in settings.devel and settings.runtests)
DEBUG = False

# Logging level used by selected log statements whose output may be useful
# for tracing field values displayed in Annalist edit/view forms.
# Suggested use is to raise level to logging.INFO when running a single named
# test, when trying to understand how values end up in a form.
TRACE_FIELD_VALUE   = logging.INFO

ALLOWED_HOSTS = []

ROOT_URLCONF = 'annalist_site.urls'

WSGI_APPLICATION = 'annalist_site.wsgi.application'

SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTOCOL", "https")

# Customize authentication backends
AUTHENTICATION_BACKENDS = (
    'django.contrib.auth.backends.ModelBackend',    # default
    'login.OAuth2CheckBackend.OAuth2CheckBackend'
    )

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'annalist',
    'login',
)

MIDDLEWARE = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            "templates"
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.contrib.auth.context_processors.auth',
                'django.template.context_processors.debug',
                'django.template.context_processors.i18n',
                'django.template.context_processors.media',
                'django.template.context_processors.static',
                'django.template.context_processors.tz',
                'django.contrib.messages.context_processors.messages',
            ],
            'debug': False,
        },
    },
]

# Database
# https://docs.djangoproject.com/en/1.6/ref/settings/#databases

# DATABASE_PATH = os.path.join(SAMPLEDATA_DIR, 'annalist_site/db.sqlite3')
# DATABASES     = {
#     'default': {
#         'ENGINE': 'django.db.backends.sqlite3',
#         'NAME':   DATABASE_PATH,
#     }
# }

# Internationalization
# https://docs.djangoproject.com/en/1.6/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.6/howto/static-files/

ADMIN_MEDIA_PREFIX = '/static/admin/'

STATIC_SEG = 'static'
STATIC_URL = '/'+STATIC_SEG+"/"

STATICFILES_DIRS = (
    ("",       SITE_SRC_ROOT+"/annalist/data/static"),
    ("images", SITE_SRC_ROOT+"/annalist/data/identity_providers/images"),
)

ANNALIST_VERSION = __version__
ANNALIST_VERSION_MSG = "Annalist version %s (common configuration)"%(ANNALIST_VERSION)

# End.



#! /usr/bin/env python -Wall
# pylint: disable=multiple-imports, ungrouped-imports

"""
Run doctests from a given set of modules in django environment.
@author: Asad
@license: BSD
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

import os, sys, glob
from sys import stderr 
import doctest

os.environ['DJANGO_SETTINGS_MODULE'] = "annalist_site.settings.runtests"
from django.test.utils import setup_test_environment

if len(sys.argv) < 2:
    n = sys.argv[0]
    print("Usage: %s path_to_test_file(s) [verbose 0 or 1]" % n, file=stderr)
    print("Example: %s appName/lib/a.py 1" % n, file=stderr)
    print("Example: %s appName/lib/ <== will search directory" % n, file=stderr)
    print("Example: %s appName/lib <== will search directory" % n, file=stderr)
    sys.exit(1)

path = sys.argv[1]
verbose = False
if len(sys.argv) == 3 and int(sys.argv[2]) == 1:
    verbose = True

ls = []
if os.path.splitext(path)[1] == '.py' and os.path.isfile(path):
    ls = [path]
elif os.path.isdir(path): 
    path = os.path.join(path, '*.py')
    ls = glob.glob(path)
else:
    print("Don't know how to deal with", path, file=stderr)
    if not os.path.exists(path):
        print("Path does not exist", file=stderr)
    sys.exit(1)


ls = [fn for fn in ls if not os.path.basename(fn).startswith('__')]
print("Finding doctests in %r"%(ls,))
results = []
for fn in ls:
    setup_test_environment()
    f = os.path.splitext(fn)[0].replace(os.sep, '.')
    m = __import__(f, {}, {}, ['*'])
    result = doctest.testmod(m, verbose=verbose)
    if verbose: 
        print(f, result, file=stderr) 
    results.append((f, result))

totalTests = 0
totalFailures = 0
for (f, result) in results:
    totalFailures += result.failed
    totalTests += result.attempted

if totalFailures > 0:
    print("Summary:", \
          totalFailures, "failures out of", \
          totalTests, "attempts",
          file=stderr)
    sys.exit(1)
else:
    print("All", totalTests, "tests passed")

# End.



"""
This module contains utilities for use in conjunction with field renderers.
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2014, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import traceback

from collections            import OrderedDict, namedtuple

from django.conf            import settings
from django.utils.html      import escape

from utils.py3porting       import is_string, to_unicode, urljoin

from annalist.exceptions    import TargetIdNotFound_Error, TargetEntityNotFound_Error
from annalist.identifiers   import RDFS, ANNAL
from annalist.util          import split_type_entity_id

from annalist.models.entitytypeinfo         import EntityTypeInfo
from annalist.models.entity                 import EntityRoot

from annalist.views.uri_builder             import (
    uri_params, uri_with_params, continuation_params
    )
from annalist.views.form_utils.fieldchoice  import FieldChoice

# -----------------------------------------------------------------------------
# Field description for doctests (avoids circular imports via FieldDescription)
# -----------------------------------------------------------------------------

class MockFieldDescription(object):
    """
    Simplified field description for local testing.
    (Somehow, can't get it working with real FieldDescription...)
    (I think it may be a problem with import cycles...)
    """

    def __init__(self, coll, recordfield):
        self._field_desc = recordfield
        return

    def __copy__(self):
        """
        Shallow copy of self.
        """
        cls = self.__class__
        result = cls.__new__(cls)
        result._field_desc = self._field_desc
        return result

    def __repr__(self):
        return repr(self._field_desc)

    def __getitem__(self, k):
        """
        Allow direct indexing to access field description fields
        """
        return self._field_desc[k]

    def get(self, name, default):
        return self._field_desc.get(name, default)

    def get_field_id(self):
        """
        Returns the field identifier
        """
        return self._field_desc['field_id']

    def get_field_name(self):
        """
        Returns form field name to be used for the described field
        """
        return self.get('field_name', self.get_field_id())

    def get_field_property_uri(self):
        """
        Returns form field property URI to be used for the described field
        """
        return self._field_desc['field_property_uri']

    def get_field_subproperty_uris(self):
        """
        Returns list of possible subproperty URIs for the described field
        """
        return self._field_desc.get('field_subproperty_uris', [])

    def get_field_value_key(self, entityvals):
        """
        Returns field value key for use in supplied entity values instance
        """
        return self.get_field_property_uri()

# -----------------------------------------------------------------------------
# Bound field
# -----------------------------------------------------------------------------

class bound_field(object):
    """
    Class representing an entity bound to a field description, 
    which can be used as input for data driven rendering of a form field.

    The field description contains information used to extract the field
    value from the entity.

    This class, and in particular its `__getattr__` method, is provided to 
    allow indirected access to an entity value fields to be performed using a 
    Django template using, e.g., "{{ field.field_value }}" (thus satisfying the 
    Django design goal that computation is handled in the python code rather 
    than the template, though in this case the computation us handled while
    rather than before the page is rendered).

    See also: http://docs.python.org/2/reference/datamodel.html#slots

    --- from annalist.views.fields.field_description import FieldDescription

    >>> coll = None
    >>> entity = EntityRoot("entityuri", "entityuri", "entitydir", "entitydir")
    >>> entity.set_id("testentity")
    >>> vals = entity.set_values({"foo": "foo_val", "bar": "bar_val"})
    >>> field_foo_desc = MockFieldDescription(coll, {"field_id": "foo_id", "field_property_uri": "foo", "field_type": "foo_type"})
    >>> field_foo = bound_field(field_foo_desc, entity)
    >>> field_foo._key == 'foo'
    True
    >>> field_foo.description['field_type'] == 'foo_type'
    True
    >>> field_foo.field_value == 'foo_val'
    True
    >>> field_bar_desc = MockFieldDescription(coll, {"field_id": "bar_id", "field_property_uri": "bar", "field_type": "bar_type"})
    >>> field_bar = bound_field(field_bar_desc, entity)
    >>> field_bar.description['field_type'] == 'bar_type'
    True
    >>> field_bar.field_value == 'bar_val'
    True
    >>> field_def_desc = MockFieldDescription(coll, {"field_id": "def_id", "field_property_uri": "def", "field_type": "def_type"})
    >>> entityvals = entity.get_values()
    >>> entityvals['entity_id']      = entity.get_id()
    >>> entityvals['entity_type_id'] = entity.get_type_id()
    >>> entityvals['entity_link']    = entity.get_url()
    >>> field_def = bound_field(field_def_desc, entity)
    >>> field_def.description['field_type'] == 'def_type'
    True
    >>> field_def.field_value == ""
    True
    >>> field_def = bound_field(field_def_desc, entity, context_extra_values={"def": "default"})
    >>> field_def.description['field_type'] == 'def_type'
    True
    >>> field_def.field_value == 'default'
    True
    >>> field_def.entity_link == 'entityuri/'
    True
    """

    __slots__ = ("_field_description", "_entityvals", "_targetvals", "_key", "_extras")

    def __init__(self, field_description, entityvals, context_extra_values=None):
        """
        Initialize a bound_field object.

        field_description       is a dictionary-like object describing a display
                                field.  See `FieldDescription` class for more details.
        entityvals              is an entity values dictionary from which a value to be 
                                rendered is obtained.  The specific field value used is 
                                defined by the combination with `field_description`.  
        context_extra_values    if supplied, a supplementary value dictionary that may be
                                probed for values that are not provided by the entity itself.  
                                Can be used to specify default values for an entity.
        """
        # if not isinstance(entityvals, dict):
        #     raise ValueError("bound_field entityvals is not dictionary (%r)"%(entityvals,))
        self._field_description = field_description
        self._entityvals        = entityvals
        self._targetvals        = None
        self._key               = field_description.get_field_property_uri()
        self._extras            = context_extra_values
        return

    def __copy__(self):
        """
        Shallow(-ish) copy of self.

        (Tried code from http://stackoverflow.com/a/15774013, but hits recursion limit)
        """
        cls = self.__class__
        result = cls.__new__(cls)
        result._field_description = self._field_description.copy()
        result._entityvals        = self._entityvals
        result._targetvals        = self._targetvals
        result._key               = self._key
        result._extras            = self._extras
        return result

    def __getattr__(self, name):
        """
        Get a bound field description attribute.  Broadly, if the attribute name is 
        "field_value" then the value corresponding to the field description is 
        retrieved from the entity, otherwise the named attribute is retrieved from 
        the field description.

        There are also a number of other special cases handled here as needed to 
        support internally-generated hyperlinks and internal system logic.
        """
        # log.info("self._key %s, __getattr__ %s"%(self._key, name))
        # log.info("self._key %s"%self._key)
        # log.info("self._entity %r"%self._entity)
        if name in ["entity_id", "entity_link", "entity_type_id", "entity_type_link"]:
            return self._entityvals.get(name, "")

        elif name == "entity_value":
            return self._entityvals
        elif name in ["field_value", "field_edit_value"]:
            return self.get_field_value()
        elif name == "field_value_key":
            return self.get_field_value_key()
        elif name == "field_value_link":
            return self.get_field_selection_link()
        elif name in ["target_value", "field_view_value"]:
            return self.get_target_value()
        elif name == "target_value_link":
            return self.get_target_link()
        elif name == "continuation_url":
            return self.get_continuation_url()
        elif name == "continuation_param":
            return self.get_continuation_param()

        elif name == "field_id":
            return self._field_description.get_field_id()
        elif name == "field_name":
            return self._field_description.get_field_name()
        elif name == "field_label":
            return self._field_description["field_label"]
        elif name == "field_help":
            return self.get_field_help_esc()
        elif name == "field_tooltip":
            return self.get_field_tooltip()
        elif name == "field_tooltip_attr":
            return self.get_field_tooltip_attr()
        elif name == "render":
            return self._field_description["field_renderer"]
        elif name == "value_mapper":
            return self._field_description["field_value_mapper"]
        elif name == "description":
            return self._field_description

        elif name == "field_value_key":
            return self._key
        elif name == "context_extra_values":
            return self._extras
        elif name == "options":
            return self.get_field_options()
        elif name == "copy":
            return self.__copy__

        # elif name == "row_field_descs":
        elif True:      # For diagnosing failed accesses...
            msg = "@@@@ Accessing bound_field.%s"%(name,)
            log.error(msg)
            log.debug("".join(traceback.format_stack()))
            assert False, msg
        return "@@bound_field.%s@@"%(name)

    def get_field_value(self):
        """
        Return field value corresponding to key from field description.
        """
        field_key = self.get_field_value_key()
        # log.debug(
        #     "@@ bound_field.get_field_value field_key %s, _entityvals %r"%
        #     (field_key, self._entityvals)
        #     )
        if hasattr(self._entityvals, "get"):
            field_val = self._entityvals.get(field_key, None)
        else:
            field_val = "@@ Cannot resolve: %s[%s]"%(self._entityvals, field_key)
        # Allow field value to be provided via `context_extra_values` if not in entity.
        # (Currently used for 'get_view_choices_field' and 'get_list_choices_field'
        # to insert current display selection.)
        if field_val is None and self._extras and self._key in self._extras:
            field_val = self._extras[self._key]
        # If no value present, use default from field definition, or blank value
        if field_val is None:
            field_val = self._field_description.get('field_default_value', None)
            if field_val is None:
                field_val = ""
        return field_val

    def get_field_value_key(self):
        """
        Return field value key used in current entity.

        This takes account of possible use of subproperties of the property URI
        specified in the field description.  If the declared property URI is not 
        present in the entity, and a subproperty URI is present, then that 
        subproperty URI is returned.  Otherwise the declared property URI is returned.
        """
        return self._field_description.get_field_value_key(self._entityvals)

    def get_field_selection_link(self):
        """
        Return a link corresponding to a field value that is a selection from 
        an enumeration of entities (or some other value with an associated link),
        or None
        """
        choices = self._field_description['field_choices']  # OrderedDict
        v       = self.field_value
        if choices and v in choices:
            return choices[v].link
        return None

    def get_field_help_esc(self):
        """
        Return help text from field description, for use as tooltip
        """
        return escape(
            self._field_description['field_help'] or 
            "@@field help for %(field_label)s@@"%self._field_description
            )

    def get_field_tooltip(self):
        """
        Return tooltip text for displaying field popup
        """
        tooltip_text_esc = escape(
            self._field_description['field_tooltip'] or 
            (self._field_description['field_help']) or
            "@@tooltip for %(field_label)s@@"%self._field_description
            )
        return tooltip_text_esc

    def get_field_tooltip_attr(self):
        """
        Return tooltip attribute for displaying field help, or blank
        """
        tooltip_text_esc = self.get_field_tooltip()
        return ''' title="%s"'''%tooltip_text_esc if tooltip_text_esc else ''

    def get_target_value(self):
        """
        Get target value of field for view display.

        This may be different from field_value if it references another entity field
        """
        targetvals = self.get_targetvals()
        # log.debug("bound_field.get_target_value: targetvals %r"%(targetvals,))
        target_key = self._field_description.get('field_ref_field', None)
        target_key = target_key and target_key.strip()
        if targetvals is not None:
            if target_key:
                log.debug("bound_field.get_target_value: target_key %s"%(target_key,))
                target_value = targetvals.get(target_key, "(@%s)"%(target_key))
            else:
                target_value = targetvals
        elif target_key:
            target_value = self._entityvals.get(target_key, "(@%s)"%(target_key))
        else:
            target_value = self.field_value
        # log.debug("bound_field.get_target_value result: %r"%(target_value,))
        return target_value

    def get_target_link(self):
        """
        Return link corresponding to target value, or None.

        The target value is treated as a relative reference relative to 
        the field_link value.  If the target value is itself an absolute URI, 
        it will be used as-is.

        If the target value is a dictionary structure created by a URIImport or 
        FileUpload field, the resulting value links to the imported data object.

        If the field is a reference to values of another type (i.e. a selection 
        field), then the field field value is used to determine the selected entity, 
        and the entity link is used as the base URI against which the target value 
        is resolved (the entity URI referencing a directory or container).
        """
        target_base  = self.get_field_selection_link() or self.entity_link
        target_value = self.get_target_value()
        # log.debug("get_target_link: base %r, value %r"%(target_base, target_value))
        if target_base and target_value:
            if isinstance(target_value, dict) and 'resource_name' in target_value:
                target_ref = target_value['resource_name']
            elif is_string(target_value):
                target_ref = target_value
            else:
                log.warning(
                    "bound_field.get_target_link: "+
                    "target_value must be URI string or URIImport structure; got %r"%
                      (target_value,)
                    )
                target_ref = None
            return urljoin(target_base, target_ref)
        return target_value

    def get_targetvals(self):
        """
        If field description is a reference to a target type entity or field, 
        return a copy of the referenced target entity, otherwise None.
        """
        # log.debug("@@ bound_field.get_targetvals: field_description %r"%(self._field_description,))
        target_type = self._field_description.get('field_ref_type',  None)
        target_key  = self._field_description.get('field_ref_field', None)
        log.debug("bound_field.get_targetvals: target_type '%s', target_key '%s'"%(target_type, target_key))
        if self._targetvals is None:
            if target_type:
                # Extract entity_id and type_id; default to type id from field descr
                field_val = self.get_field_value()
                log.debug("field_val: %s"%(field_val,))
                type_id, entity_id = split_type_entity_id(self.get_field_value(), target_type)
                log.debug("bound_field.get_targetvals: type_id %s, entity_id %s"%(type_id, entity_id))
                # Get entity type info
                coll     = self._field_description._collection
                typeinfo = EntityTypeInfo(coll, type_id)
                # Check access permission, assuming user has "VIEW" permission in collection
                # This is primarily to prevent a loophole for accessing user account details
                #@@TODO: pass actual user permissions in to bound_field or field description 
                #        or extra params
                user_permissions    = ["VIEW"]
                req_permissions_map = typeinfo.get_entity_permissions_map(entity_id)
                req_permissions     = list(set( req_permissions_map[a] for a in ["view", "list"] ))
                if all([ p in user_permissions for p in req_permissions]):
                    if entity_id is None or entity_id == "":
                        raise TargetIdNotFound_Error(value=(typeinfo.type_id, self._field_description["field_name"]))
                    targetentity = typeinfo.get_entity(entity_id)
                    if targetentity is None:
                        raise TargetEntityNotFound_Error(value=(target_type, entity_id))
                    targetentity = typeinfo.get_entity_implied_values(targetentity)
                    self._targetvals = get_entity_values(typeinfo, targetentity)
                    log.debug("bound_field.get_targetvals: %r"%(self._targetvals,))
                else:
                    log.warning(
                        "bound_field.get_targetvals: target value type %s requires %r permissions"%
                        (target_type, req_permissions)
                        )
        log.debug("bound_field.get_targetvals: targetvals %r"%(self._targetvals,))
        return self._targetvals

    def get_link_continuation(self, link):
        """
        Return supplied base link with continuation parameter appended
        (if the link value is defined - i.e. not None or empty).
        """
        if link:
            link += self.get_continuation_param()
        return link

    def get_continuation_param(self):
        """
        Generate continuation parameter string for return back to the current request page
        """
        cparam = ""
        chere  = self.get_continuation_url()
        if chere:
            cparam = uri_params({'continuation_url': chere})
        return cparam

    def get_continuation_url(self):
        """
        Generate continuation URL for return back to the current request page
        """
        chere = ""
        if self._extras is None:
            log.warning("bound_field.get_continuation_url() - no extra context provided")
        else:
            requrl = self._extras.get("request_url", "")
            if requrl:
                chere  = uri_with_params(requrl, continuation_params(self._extras))
        # log.debug('bound_field.get_continuation_url %s'%(chere,))
        return chere

    def get_field_options(self):
        """
        Returns list of selectable options for the current field

        Note: in Python3, OrderedDict.values() returns a view, not a list.
        """
        options = self._field_description['field_choices']      # OrderedDict
        options = ( list(options.values()) if options is not None else 
                    [ FieldChoice('', label="(no options)") ]
                  )
        return options

    def __getitem__(self, name):
        return self.__getattr__(name)

    def __iter__(self):
        """
        Implement iterator protocol, returning accessible value keys.
        """
        yield "entity_id"
        yield "entity_type_id"
        yield "entity_link"
        yield "entity_type_link"
        yield "field_edit_value"
        yield "field_view_value"
        yield "field_value"
        yield "field_value_key"
        yield "field_value_link"
        yield "target_value"
        yield "target_value_link"
        yield "continuation_url"
        yield "continuation_param"
        yield "description"
        yield "options"
        # Direct access to selected field descrption attributes 
        yield "field_id"
        yield "field_name"
        yield "field_label"
        yield "field_help"
        yield "field_tooltip"
        yield "field_tooltip_attr"
        return

    def as_dict(self):
        return dict( # self._field_description.items(), 
            entity=dict(self._entityvals.items()), 
            field_value=self.field_value, 
            context_extra_values=self._extras, 
            key=self._key
            )

    def __repr__(self):
        return self.fullrepr()

    def __str__(self):
        return self.shortrepr()

    def __unicode__(self):
        return self.shortrepr()

    def shortrepr(self):
        return (
            "bound_field(\n"+
            "  { 'key': %r\n"%(self._key,)+
            "  , 'val': %r\n"%(self.field_value,)+
            "  , 'field_description': %r\n"%(self._field_description,)+
            "  })")

    def fullrepr(self):
        try:
            field_view_value = self.field_view_value
        except Exception as e:
            field_view_value = str(e)
        return (
            ( "bound_field(\n"+
              "  { 'key': %r\n"+
              "  , 'field_edit_value': %r\n"+
              "  , 'field_view_value': %r\n"+
              "  , 'description': %r\n"+
              "  , 'entity_vals': %r\n"+
              "  })\n"
            )%( self._key
              , self.field_edit_value
              , field_view_value 
              , self._field_description
              , dict(self._entityvals.items())
              )
            )

    def htmlrepr(self):
        return (
            "<ul>"+
            "<li>key: %s</li>"%(self._key,)+
            "<li>val: %s</li>"%(self.field_value,)+
            "<li>field_description: %r</li>"%(self._field_description,)+
            "</ul>"
            )

# -----------------------------------------------------------------------------
# Helper functions
# -----------------------------------------------------------------------------

def get_entity_values(typeinfo=None, entity=None, entity_id=None, action="view"):
    """
    Returns an entity values dictionary for a supplied entity, suitable for
    use with a bound_field object.
    """
    if not entity_id:
        entity_id = entity.get_id()
    type_id    = entity.get_type_id()
    entityvals = entity.get_values().copy()
    entityvals['entity_id']        = entity_id
    entityvals['entity_link']      = entity.get_view_url_path()
    # log.info("type_id %s"%(type_id))
    entityvals['entity_type_id']   = type_id
    if action == "copy":
        # Allocate new entity Id and lose values based on original Id 
        # when performing a copy operation.
        # entity_id = typeinfo.entityclass.allocate_new_id(typeinfo.entityparent)
        # log.info("@@ copy new entity_id %s"%entity_id)
        # entityvals['entity_id'] = entity_id
        entityvals.pop('entity_link', None)
        entityvals[ANNAL.CURIE.id] = entity_id
        entityvals.pop(ANNAL.CURIE.uri, None)
    if typeinfo and typeinfo.recordtype:
        entityvals['entity_type_link'] = typeinfo.recordtype.get_view_url_path()
    # else:
    #     log.error("get_entity_values: No recordtype: typeinfo %r"%(typeinfo,))
    #     log.error("get_entity_values: No recordtype: entity id %s/%s"%(type_id,entity_id))
    #     log.error("get_entity_values: No recordtype: entity %r"%(entity,))
    #     traceback.print_stack()
    return entityvals

if __name__ == "__main__":
    import doctest
    doctest.testmod()

# End.



"""
Annalist general information entity (used for displaying arbitrary information)
"""

from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2019, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

import logging
log = logging.getLogger(__name__)

import os
import os.path
import shutil

from annalist                   import layout
from annalist.exceptions        import Annalist_Error
from annalist.identifiers       import ANNAL, RDFS, OWL
from annalist                   import util
from annalist.models.entitydata import EntityData

class RecordInfo(EntityData):

    _entitytype     = ANNAL.CURIE.Information
    _entitytypeid   = layout.INFO_TYPEID
    _entityroot     = layout.COLL_INFO_PATH
    _entityview     = layout.COLL_INFO_VIEW
    _entityfile     = layout.INFO_META_FILE

    def __init__(self, parent, info_id):
        """
        Initialize a new RecordInfo object, without metadata (yet).

        parent      is the parent collection in which the information is defined.
        info_id     the local identifier for the information.
        """
        super(RecordInfo, self).__init__(parent, info_id)
        self._parent = parent
        # log.debug("RecordInfo %s: dir %s"%(info_id, self._entitydir))
        return

# End.
